function [X, iter, Outputs] = PCalScalable(g, M, param)
% Minimizes Sum_k ||X{k,k}||_1+||X{k,k+1}||_1    s.t.   g = A * X(:) , 
%       [X{k,k} X{k,k+1};X{k,k+1}' X{k+1,k+1}] is positive semidefinite for
%       all k
%
% ADMM formulation:
% Sum_k Ind_pos.sem.def(V{k}) + ||U{k,k}||_1+||U{k,k+1}||_1   s.t. g = A * X(:)
%                                            V{k} = [X{k,k} X{k,k+1};X{k,k+1}' X{k+1,k+1}]
%                                            U{k,k} = X{k,k}, U{k,k+1} = X{k,k+1}
% 
% X : LxL cells (only 2L non-empty) with NxN matrices
% g : LxL cells (only 2L non-empty) with mx1 vectors
%
% Ind_pos.sem.def() is the indicator function for hermitian positive semi
% definite matrices
%
% param.mu : ADMM parameter for V=X, U=X and g=A * X(:)
% param.N : signal size
% param.M : measurement size
% param.L : number of experiments
% param.Nit : number of iterations
% param.convergence_err : the limit of error between 
%                           consecutive updates for convergence
% param.timelimit : the limit for execution time in minutes, when reached
%                   simualtion is paused and all data is saved in the param.resumefile
% param.resume : flag to indicate if an earlier simulation will be resumed
% param.resumefile : the file to load/save the data to resume/pause the
%                    simulation
tic;
% --- Useful functions

Nit = param.max_iter;
timelimit=param.timelimit;
N=param.N;
L=param.L;

k_first = 1;
if(any(strcmp(fieldnames(param), 'resume')) && param.resume==1)
    Nit2=Nit;
    timelimit2=timelimit;
    load(param.resumefile);
    Nit=Nit2;
    timelimit=timelimit2;    
    clear Nit2 timelimit2;
else
    A = @(X) forward_op(X, M);
    At = @(y) adjoint_op(y, M);
    AtoA = @(X) adjoint_o_forward(X, M);
    Afun_eig = @(x) reshape(AtoA(reshape(x, N, N)), N^2, 1);

    opts.isreal = false;
    M_scale = eigs(Afun_eig,N^2,1,'LM',opts);
    mu1 = param.mu;
    mu2 = param.mu;
    mu3 = (mu1+mu2)*100/M_scale;
    Afun = @(x) reshape(mu3*AtoA(reshape(x, N, N)), N^2, 1) + (2*mu1+mu2)*x;

    convergence_err=param.convergence_err;

    show_corr =0;
    if(any(strcmp(fieldnames(param), 'x_org')))
        x_org = param.x_org;
        show_corr =1;
    end

    J_it = zeros(Nit, 5);
    X = cell(L,L);
    V = cell(L,1);
    U = cell(L,L);
    D1 = cell(L,1);
    D2 = cell(L,L);
    d3 = cell(L,L);
    AX = cell(L,L);
    for el=1:L
    X{el,el} = At(g{el,el})/M_scale;
    X{el, mod(el,L)+1}  = At(g{el, mod(el,L)+1})/M_scale;
    D1{el,1} = zeros(2*size(X{el,el}));
    D2{el,el} = zeros(size(X{el,el}));
    D2{el, mod(el,L)+1} = zeros(size(X{el, mod(el,L)+1}));
    d3{el,el} = zeros(size(g{el,el}));
    d3{el, mod(el,L)+1} = zeros(size(g{el, mod(el,L)+1}));
    V{el,1} = zeros(2*size(X{el,el}));
    U{el,el} = zeros(size(X{el,el}));
    U{el, mod(el,L)+1} = zeros(size(X{el, mod(el,L)+1}));
    end
    update_err = inf;
    
    corr_reached=false;
end

iter=k_first;
while (((iter<=Nit)&&(update_err>convergence_err)&&(~corr_reached)&&((toc/60)<(timelimit*0.95))))
 
    % Vk = mu1/2 || [Xk,k Xk,k+1;Xk,k+1' Xk+1,k+1] - Vk - D1k ||_2^2 s.t. Vk >= 0
    %   =  Positive eigenvalues of  1/2 ([Xk,k Xk,k+1;Xk,k+1' Xk+1,k+1] - D1k - [Xk,k Xk,k+1;Xk,k+1' Xk+1,k+1]' + D1k')_+ 
    num_eigs=0;
    for el=1:L
        temp = [X{el,el} X{el,mod(el,L)+1} ; X{el,mod(el,L)+1}' X{mod(el,L)+1,mod(el,L)+1}];
        V{el,1} = (temp - D1{el,1} + temp' - D1{el,1}')/2;
        [C,S]=eig(V{el,1});
        S = real(diag(S)); ind_Snz = (S>1e-14);
        S(~ind_Snz)=0;
        V{el,1} = C(:,ind_Snz)*diag(S(ind_Snz))*C(:,ind_Snz)';
        num_eigs = max(sum(ind_Snz),num_eigs);
    end 
    
    % Uk,l = argmin  || Uk,l ||_1 + mu2/2 || Xk,l - Uk,l - D2k,l ||_2^2
    %   = soft(Xk,l-D2k,l , 1/mu2)
    for el=1:L
        U{el,el} = soft(X{el,el} - D2{el,el}, 1/mu2);
        U{el, mod(el,L)+1} = soft(X{el, mod(el,L)+1} - D2{el, mod(el,L)+1}, 1/mu2);
    end
    
    % Xk,k = argmin mu1/2 || Xk,k - Vk(1,1) - D1k(1,1) ||_2^2 + 
    %               mu1/2 || Xk,k - Vk-1(2,2) - D1k-1(2,2) ||_2^2 + mu2/2 || Xk,k - Uk,k - D2k,k ||_2^2 
    %            + mu3/2 || zk,k - A * X(:)k,k - d3k,k ||_2^2
    %   = ( (2*mu1+mu2) I + mu3 A'*A )^-1 [ mu3 A' *(zk,k-d3k,k) + mu1( Vk(1,1)+D1k(1,1) ) 
    %                                     + mu1( Vk-1(2,2)+D1k-1(2,2) ) + mu2( Uk,k+D2k,k )]
    X_old = X;
    
    update_err=0;
    for el=1:L
        a = mu1*(V{el,1}(1:N,1:N)+D1{el,1}(1:N,1:N)) + mu1*(V{mod(el-2,L)+1,1}(N+1:2*N,N+1:2*N)+D1{mod(el-2,L)+1,1}(N+1:2*N,N+1:2*N))...
          + mu2*(U{el,el}+D2{el,el}) + mu3*(At((g{el,el}-d3{el,el})));
    
        [m, fl] = pcg(Afun, a(:), 10^-8, 1000, [], [], X{el,el}(:));
        X{el,el} = reshape(m, size(X{el,el}));
        if(fl)
            disp('Inversion of I + A"*A not converged');
        end
        update_err = max(update_err,norm(X{el,el}(:)-X_old{el,el}(:))/norm(X_old{el,el}(:)));
        AX{el,el} = A(X{el,el});
        
        a = mu1*(V{el,1}(1:N,N+1:2*N)+D1{el,1}(1:N,N+1:2*N)+V{el,1}(N+1:2*N,1:N)'+D1{el,1}(N+1:2*N,1:N)')...
          + mu2*(U{el,mod(el,L)+1}+D2{el,mod(el,L)+1}) + mu3*(At((g{el,mod(el,L)+1}-d3{el,mod(el,L)+1})));
    
        [m, fl] = pcg(Afun, a(:), 10^-8, 1000, [], [], X{el,mod(el,L)+1}(:));
        X{el,mod(el,L)+1} = reshape(m, size(X{el,mod(el,L)+1}));
        if(fl)
            disp('Inversion of I + A"*A not converged');
        end
        update_err = max(update_err,norm(X{el,mod(el,L)+1}(:)-X_old{el,mod(el,L)+1}(:))/norm(X_old{el,mod(el,L)+1}(:)));
        AX{el,mod(el,L)+1} = A(X{el,mod(el,L)+1});
    end
    
    
    % Update D1, D2, d3
    for el=1:L
        D1{el,1} = D1{el,1} - ([X{el,el} X{el,mod(el,L)+1} ; X{el,mod(el,L)+1}' X{mod(el,L)+1,mod(el,L)+1}]-V{el,1});
        D2{el,el} = D2{el,el} - (X{el,el}-U{el,el});
        D2{el,mod(el,L)+1} = D2{el,mod(el,L)+1} - (X{el,mod(el,L)+1}-U{el,mod(el,L)+1});
        d3{el,el} = d3{el,el} - (g{el,el} - AX{el,el});
        d3{el,mod(el,L)+1} = d3{el,mod(el,L)+1} - (g{el,mod(el,L)+1} - AX{el,mod(el,L)+1});
    end
    
    J_it(iter, :)=0;
    for el=1:L
        J_it(iter, 1) = J_it(iter, 1) + sum(real(trace(V{el,1})))/2;
        J_it(iter, 2) = J_it(iter, 2) + sum(abs(U{el,el}(:))) + sum(abs(U{mod(el,L)+1,el}(:)));
        temp = [X{el,el} X{el, mod(el,L)+1}; X{el, mod(el,L)+1}' X{mod(el,L)+1,mod(el,L)+1}];
        J_it(iter, 3) = J_it(iter, 3) + sum(abs(temp(:) - V{el,1}(:)).^2);  
        J_it(iter, 4) = J_it(iter, 4) + sum(abs(X{el,el}(:) - U{el,el}(:)).^2) + sum(abs(X{el,mod(el,L)+1}(:) - U{el,mod(el,L)+1}(:)).^2);  
        J_it(iter, 5) = J_it(iter, 5) + sum(abs(g{el,el}(:) - AX{el,el}(:)).^2) + sum(abs(g{el,mod(el,L)+1}(:) - AX{el,mod(el,L)+1}(:)).^2);
    end
    J_it(iter, 6) = update_err;
    
    if(mod(iter,10)==0)
        if(show_corr)
            ph=1;
            for el=L:-1:1
                temp = [X{el,el} X{el,mod(el,L)+1} ; X{el,mod(el,L)+1}' X{mod(el,L)+1,mod(el,L)+1}];
                [xrect, lbd] = eigs(temp,1);
                xrect=xrect(:)*sqrt(lbd);
                newph=mean(xrect(N+1:2*N))/abs(mean(xrect(N+1:2*N)));
                xrec(:,el) = xrect(1:N)*ph/newph;
                ph = mean(xrec(:,el))/abs(mean(xrec(:,el)));
            end          
            x_corr = abs(sum(xrec(:)'*x_org(:)))/norm(xrec(:))/norm(x_org(:));
            corr_reached = x_corr>0.9999;
            disp(['It=' num2str(iter) ', Corr=' num2str(x_corr) ', Diff=' num2str(update_err)...
              ', V_rank=' num2str(num_eigs) ', Err1=' num2str(J_it(iter, 3))...
              ', Err2=' num2str(J_it(iter, 4)) ', Err3=' num2str(J_it(iter, 5))]);
        else
            
                disp(['It=' num2str(iter) ', Diff=' num2str(update_err)...
              ', V_rank=' num2str(num_eigs) ', Err1=' num2str(J_it(iter, 3))...
              ', Err2=' num2str(J_it(iter, 4)) ', Err3=' num2str(J_it(iter, 5))]);
        end
    end
    
    iter=iter+1;
end

if(show_corr)
    ph=1;
    for el=L:-1:1
        temp = [X{el,el} X{el,mod(el,L)+1} ; X{el,mod(el,L)+1}' X{mod(el,L)+1,mod(el,L)+1}];
        [xrect, lbd] = eigs(temp,1);
        xrect=xrect(:)*sqrt(lbd);
        newph=mean(xrect(N+1:2*N))/abs(mean(xrect(N+1:2*N)));
        xrec(:,el) = xrect(1:N)*ph/newph;
        ph = mean(xrec(:,el))/abs(mean(xrec(:,el)));
    end  
    x_corr = abs(sum(xrec(:)'*x_org(:)))/norm(xrec(:))/norm(x_org(:));

    disp(['It=' num2str(iter-1) ', Corr=' num2str(x_corr) ', Diff=' num2str(update_err)...
          ', V_rank=' num2str(num_eigs) ', Err1=' num2str(J_it(iter-1, 3))...
          ', Err2=' num2str(J_it(iter-1, 4)) ', Err3=' num2str(J_it(iter-1, 5))]);
    else
    disp(['It=' num2str(iter-1) ', Diff=' num2str(update_err)...
          ', V_rank=' num2str(num_eigs) ', Err1=' num2str(J_it(iter-1, 3))...
          ', Err2=' num2str(J_it(iter-1, 4)) ', Err3=' num2str(J_it(iter-1, 5))]);
end

Outputs.toberesumed=0;
if(any(strcmp(fieldnames(param), 'resumefile')) && ((iter<=Nit)&&(update_err>convergence_err)&&(~corr_reached)))
    k_first = iter; %#ok<NASGU>
    save(param.resumefile);
    Outputs.toberesumed=1;
    disp(['Saving state to ' param.resumefile ' to resume later']);
end

Outputs.V = V;
Outputs.U = U;
Outputs.J_it = J_it;
end

function y = soft(x,T)
%% SOFT THRESHOLD
% y = soft(x,T)
% x : input data
% T : threshold

y = max(1-T./(abs(x)+eps),0) .* x;
end

function y = forward_op(X, M)

m = size(M,1);
y = zeros(m,1);
y(:,1) = sum((M*X).*conj(M),2).';

end

function X = adjoint_op(y, M)

X = M'*diag(y(:,1))*M;

end

function X = adjoint_o_forward(X, M)

%
y = sum((M*X).*conj(M),2).';
X = M'*diag(y)*M;

end